"""Helper functions for the NAPALM base."""
# std libs
import os
import re
import sys
import itertools
import logging
from collections import Iterable

# third party libs
import jinja2
import textfsm
from netaddr import EUI
from netaddr import mac_unix
from netaddr import IPAddress
from ciscoconfparse import CiscoConfParse

# local modules
import napalm.base.exceptions
from napalm.base import constants
from napalm.base.utils.jinja_filters import CustomJinjaFilters
from napalm.base.canonical_map import base_interfaces, reverse_mapping

# -------------------------------------------------------------------
# Functional Global
# -------------------------------------------------------------------
logger = logging.getLogger(__name__)


# -------------------------------------------------------------------
# helper classes -- will not be exported
# -------------------------------------------------------------------
class _MACFormat(mac_unix):
    pass


_MACFormat.word_fmt = "%.2X"


# -------------------------------------------------------------------
# callable helpers
# -------------------------------------------------------------------
def load_template(
    cls,
    template_name,
    template_source=None,
    template_path=None,
    openconfig=False,
    jinja_filters={},
    **template_vars,
):
    try:
        search_path = []
        if isinstance(template_source, str):
            template = jinja2.Template(template_source)
        else:
            if template_path is not None:
                if (
                    isinstance(template_path, str)
                    and os.path.isdir(template_path)
                    and os.path.isabs(template_path)
                ):
                    # append driver name at the end of the custom path
                    search_path.append(
                        os.path.join(template_path, cls.__module__.split(".")[-1])
                    )
                else:
                    raise IOError(
                        "Template path does not exist: {}".format(template_path)
                    )
            else:
                # Search modules for template paths
                search_path = [
                    os.path.dirname(os.path.abspath(sys.modules[c.__module__].__file__))
                    for c in cls.__class__.mro()
                    if c is not object
                ]

            if openconfig:
                search_path = ["{}/oc_templates".format(s) for s in search_path]
            else:
                search_path = ["{}/templates".format(s) for s in search_path]

            loader = jinja2.FileSystemLoader(search_path)
            environment = jinja2.Environment(loader=loader)

            for filter_name, filter_function in itertools.chain(
                CustomJinjaFilters.filters().items(), jinja_filters.items()
            ):
                environment.filters[filter_name] = filter_function

            template = environment.get_template(
                "{template_name}.j2".format(template_name=template_name)
            )
        configuration = template.render(**template_vars)
    except jinja2.exceptions.TemplateNotFound:
        raise napalm.base.exceptions.TemplateNotImplemented(
            "Config template {template_name}.j2 not found in search path: {sp}".format(
                template_name=template_name, sp=search_path
            )
        )
    except (
        jinja2.exceptions.UndefinedError,
        jinja2.exceptions.TemplateSyntaxError,
    ) as jinjaerr:
        raise napalm.base.exceptions.TemplateRenderException(
            "Unable to render the Jinja config template {template_name}: {error}".format(
                template_name=template_name, error=str(jinjaerr)
            )
        )
    return cls.load_merge_candidate(config=configuration)


def cisco_conf_parse_parents(parent, child, config):
    """
    Use CiscoConfParse to find parent lines that contain a specific child line.

    :param parent: The parent line to search for
    :param child:  The child line required under the given parent
    :param config: The device running/startup config
    """
    if type(config) == str:
        config = config.splitlines()
    parse = CiscoConfParse(config)
    cfg_obj = parse.find_parents_w_child(parent, child)
    return cfg_obj


def cisco_conf_parse_objects(cfg_section, config):
    """
    Use CiscoConfParse to find and return a section of Cisco IOS config.
    Similar to "show run | section <cfg_section>"

    :param cfg_section: The section of the config to return eg. "router bgp"
    :param config: The running/startup config of the device to parse
    """
    return_config = []
    if type(config) is str:
        config = config.splitlines()
    parse = CiscoConfParse(config)
    cfg_obj = parse.find_objects(cfg_section)
    for parent in cfg_obj:
        return_config.append(parent.text)
        for child in parent.all_children:
            return_config.append(child.text)
    return return_config


def regex_find_txt(pattern, text, default=""):
    """""
    RegEx search for pattern in text. Will try to match the data type of the "default" value
    or return the default value if no match is found.
    This is to parse IOS config like below:
    regex_find_txt(r"remote-as (65000)", "neighbor 10.0.0.1 remote-as 65000", default=0)
    RETURNS: 65001

    :param pattern: RegEx pattern to match on
    :param text: String of text ot search for "pattern" in
    :param default="": Default value and type to return on error
    """
    text = str(text)
    value = re.findall(pattern, text)
    try:
        if not value:
            logger.error("No Regex match found for pattern: %s" % (str(pattern)))
            raise Exception("No Regex match found for pattern: %s" % (str(pattern)))
        if not isinstance(value, type(default)):
            if isinstance(value, list) and len(value) == 1:
                value = value[0]
            value = type(default)(value)
    except Exception as regexFindTxtErr01:  # in case of any exception, returns default
        logger.error(
            'errorCode="regexFindTxtErr01" in napalm.base.helpers with systemMessage="%s"\
                 message="Error while attempting to find regex pattern, \
                      default to empty string"'
            % (regexFindTxtErr01)
        )
        value = default
    return value


def textfsm_extractor(cls, template_name, raw_text):
    """
    Applies a TextFSM template over a raw text and return the matching table.

    Main usage of this method will be to extract data form a non-structured output
    from a network device and return the values in a table format.

    :param cls: Instance of the driver class
    :param template_name: Specifies the name of the template to be used
    :param raw_text: Text output as the devices prompts on the CLI
    :return: table-like list of entries
    """
    textfsm_data = list()
    fsm_handler = None
    for c in cls.__class__.mro():
        if c is object:
            continue
        current_dir = os.path.dirname(
            os.path.abspath(sys.modules[c.__module__].__file__)
        )
        template_dir_path = "{current_dir}/utils/textfsm_templates".format(
            current_dir=current_dir
        )
        template_path = "{template_dir_path}/{template_name}.tpl".format(
            template_dir_path=template_dir_path, template_name=template_name
        )

        try:
            with open(template_path) as f:
                fsm_handler = textfsm.TextFSM(f)

                for obj in fsm_handler.ParseText(raw_text):
                    entry = {}
                    for index, entry_value in enumerate(obj):
                        entry[fsm_handler.header[index].lower()] = entry_value
                    textfsm_data.append(entry)

                return textfsm_data
        except IOError as textfsmExtractorErr01:  # Template not present in this class
            logger.error(
                'errorCode="textfsmExtractorErr01" in napalm.base.helpers with systemMessage="%s"\
                message="Error while attempting to apply a textfsm template to  \
                format the output returned from the device,\
                continuing loop..."'
                % (textfsmExtractorErr01)
            )
            continue  # Continue up the MRO
        except textfsm.TextFSMTemplateError as tfte:
            logging.error(
                "Wrong format of TextFSM template {template_name}: {error}".format(
                    template_name=template_name, error=str(tfte)
                )
            )
            raise napalm.base.exceptions.TemplateRenderException(
                "Wrong format of TextFSM template {template_name}: {error}".format(
                    template_name=template_name, error=str(tfte)
                )
            )

    raise napalm.base.exceptions.TemplateNotImplemented(
        "TextFSM template {template_name}.tpl is not defined under {path}".format(
            template_name=template_name, path=template_dir_path
        )
    )


def find_txt(xml_tree, path, default="", namespaces=None):
    """
    Extracts the text value from an XML tree, using XPath.
    In case of error, will return a default value.

    :param xml_tree:   the XML Tree object. Assumed is <type 'lxml.etree._Element'>.
    :param path:       XPath to be applied, in order to extract the desired data.
    :param default:    Value to be returned in case of error.
    :param namespaces: prefix-namespace mappings to process XPath
    :return: a str value.
    """
    value = ""
    try:
        xpath_applied = xml_tree.xpath(
            path, namespaces=namespaces
        )  # will consider the first match only
        xpath_length = len(xpath_applied)  # get a count of items in XML tree
        if xpath_length and xpath_applied[0] is not None:
            xpath_result = xpath_applied[0]
            if isinstance(xpath_result, type(xml_tree)):
                value = xpath_result.text.strip()
            else:
                value = xpath_result
        else:
            if xpath_applied == "":
                logger.debug(
                    "Unable to find the specified-text-element/XML path: %s in  \
                        the XML tree provided. Total Items in XML tree: %d "
                    % (path, xpath_length)
                )
    except Exception as findTxtErr01:  # in case of any exception, returns default
        logger.error(findTxtErr01)
        value = default
    return str(value)


def convert(to, who, default=""):
    """
    Converts data to a specific datatype.
    In case of error, will return a default value.

    :param to:      datatype to be casted to.
    :param who:     value to cast.
    :param default: value to return in case of error.
    :return: a str value.
    """
    if who is None:
        return default
    try:
        return to(who)
    except:  # noqa
        return default


def mac(raw):
    """
    Converts a raw string to a standardised MAC Address EUI Format.

    :param raw: the raw string containing the value of the MAC Address
    :return: a string with the MAC Address in EUI format

    Example:

    .. code-block:: python

        >>> mac('0123.4567.89ab')
        u'01:23:45:67:89:AB'

    Some vendors like Cisco return MAC addresses like a9:c5:2e:7b:6: which is not entirely valid
    (with respect to EUI48 or EUI64 standards). Therefore we need to stuff with trailing zeros

    Example
    >>> mac('a9:c5:2e:7b:6:')
    u'A9:C5:2E:7B:60:00'

    If Cisco or other obscure vendors use their own standards, will throw an error and we can fix
    later, however, still works with weird formats like:

    >>> mac('123.4567.89ab')
    u'01:23:45:67:89:AB'
    >>> mac('23.4567.89ab')
    u'00:23:45:67:89:AB'
    """
    if raw.endswith(":"):
        flat_raw = raw.replace(":", "")
        raw = "{flat_raw}{zeros_stuffed}".format(
            flat_raw=flat_raw, zeros_stuffed="0" * (12 - len(flat_raw))
        )
    return str(EUI(raw, dialect=_MACFormat))


def ip(addr, version=None):
    """
    Converts a raw string to a valid IP address. Optional version argument will detect that \
    object matches specified version.

    Motivation: the groups of the IP addreses may contain leading zeros. IPv6 addresses can \
    contain sometimes uppercase characters. E.g.: 2001:0dB8:85a3:0000:0000:8A2e:0370:7334 has \
    the same logical value as 2001:db8:85a3::8a2e:370:7334. However, their values as strings are \
    not the same.

    :param raw: the raw string containing the value of the IP Address
    :param version: (optional) insist on a specific IP address version.
    :type version: int.
    :return: a string containing the IP Address in a standard format (no leading zeros, \
    zeros-grouping, lowercase)

    Example:

    .. code-block:: python

        >>> ip('2001:0dB8:85a3:0000:0000:8A2e:0370:7334')
        u'2001:db8:85a3::8a2e:370:7334'
    """
    addr_obj = IPAddress(addr)
    if version and addr_obj.version != version:
        raise ValueError("{} is not an ipv{} address".format(addr, version))
    return str(addr_obj)


def as_number(as_number_val):
    """Convert AS Number to standardized asplain notation as an integer."""
    as_number_str = str(as_number_val)
    if "." in as_number_str:
        big, little = as_number_str.split(".")
        return (int(big) << 16) + int(little)
    else:
        return int(as_number_str)


def split_interface(intf_name):
    """Split an interface name based on first digit, slash, or space match."""
    head = intf_name.rstrip(r"/\0123456789. ")
    tail = intf_name[len(head) :].lstrip()
    return (head, tail)


def canonical_interface_name(interface, addl_name_map=None):
    """Function to return an interface's canonical name (fully expanded name).

    Use of explicit matches used to indicate a clear understanding on any potential
    match. Regex and other looser matching methods were not implmented to avoid false
    positive matches. As an example, it would make sense to do "[P|p][O|o]" which would
    incorrectly match PO = POS and Po = Port-channel, leading to a false positive, not
    easily troubleshot, found, or known.

    :param interface: The interface you are attempting to expand.
    :param addl_name_map (optional): A dict containing key/value pairs that updates
    the base mapping. Used if an OS has specific differences. e.g. {"Po": "PortChannel"} vs
    {"Po": "Port-Channel"}
    """

    name_map = {}
    name_map.update(base_interfaces)
    interface_type, interface_number = split_interface(interface)

    if isinstance(addl_name_map, dict):
        name_map.update(addl_name_map)
    # check in dict for mapping
    if name_map.get(interface_type):
        long_int = name_map.get(interface_type)
        return long_int + str(interface_number)
    # if nothing matched, return the original name
    else:
        return interface


def abbreviated_interface_name(interface, addl_name_map=None, addl_reverse_map=None):
    """Function to return an abbreviated representation of the interface name.

    :param interface: The interface you are attempting to abbreviate.
    :param addl_name_map (optional): A dict containing key/value pairs that updates
    the base mapping. Used if an OS has specific differences. e.g. {"Po": "PortChannel"} vs
    {"Po": "Port-Channel"}
    :param addl_reverse_map (optional): A dict containing key/value pairs that updates
    the reverse mapping. Used if an OS has specific differences. e.g. {"PortChannel": "Po"} vs
    {"PortChannel": "po"}
    """

    name_map = {}
    name_map.update(base_interfaces)
    interface_type, interface_number = split_interface(interface)

    if isinstance(addl_name_map, dict):
        name_map.update(addl_name_map)

    rev_name_map = {}
    rev_name_map.update(reverse_mapping)

    if isinstance(addl_reverse_map, dict):
        rev_name_map.update(addl_reverse_map)

    # Try to ensure canonical type.
    if name_map.get(interface_type):
        canonical_type = name_map.get(interface_type)
    else:
        canonical_type = interface_type

    try:
        abbreviated_name = rev_name_map[canonical_type] + str(interface_number)
        return abbreviated_name
    except KeyError:
        pass

    # If abbreviated name lookup fails, return original name
    return interface


def transform_lldp_capab(capabilities):
    if capabilities and isinstance(capabilities, str):
        capabilities = capabilities.strip().lower().split(",")
        return sorted(
            [constants.LLDP_CAPAB_TRANFORM_TABLE[c.strip()] for c in capabilities]
        )
    else:
        return []


def generate_regex_or(filters):
    """
    Build a regular expression logical-or from a list/tuple of regex patterns.

    This allows a single regular expression operation to be used in contexts when a loop
    and multiple patterns would otherwise be necessary.

    For example, (pattern1|pattern2|pattern3)

    Return the pattern.
    """
    if isinstance(filters, str) or not isinstance(filters, Iterable):
        raise ValueError("filters argument must be an iterable, but can't be a string.")

    return_pattern = r"("
    for pattern in filters:
        return_pattern += rf"{pattern}|"
    return_pattern += r")"
    return return_pattern


def sanitize_config(config, filters):
    """
    Given a list of filters, remove sensitive data from the provided config.
    """
    for filter_, replace in filters.items():
        config = re.sub(filter_, replace, config, flags=re.M)
    return config


def sanitize_configs(configs, filters):
    """
    Apply sanitize_config on the dictionary of configs typically returned by
    the get_config method.
    """
    for cfg_name, config in configs.items():
        if config.strip():
            configs[cfg_name] = sanitize_config(config, filters)
    return configs



# Based on:
# https://code.getnoc.com/noc/noc/blob/6f3db2a6e4b1ece77aaf4c4c98413e35ff64643a/sa/profiles/Arista/EOS/get_lldp_neighbors.py#L76-79
LLDP_CAPAB_TRANFORM_TABLE = {
    "other": "other",
    "repeater": "repeater",
    "bridge": "bridge",
    "wlanaccesspoint": "wlan-access-point",
    "router": "router",
    "telephone": "telephone",
    "docsis": "docsis-cable-device",
    "station": "station",
    "stationonly": "station",
}



"""
Fork of pynxos library from network to code and mzbenami

Reimplemented by ktbyers to support XML-RPC in addition to JSON-RPC
"""

from __future__ import unicode_literals


class NXAPIError(Exception):
    """Generic NXAPI exception."""

    pass


class NXAPICommandError(NXAPIError):
    def __init__(self, command, message):
        self.command = command
        self.message = message

    def __repr__(self):
        return 'The command "{}" gave the error "{}".'.format(
            self.command, self.message
        )

    __str__ = __repr__


class NXAPIConnectionError(NXAPIError):
    """HTTP Post Connection Error."""

    pass


class NXAPIAuthError(NXAPIError):
    """HTTP Post Authentication Error."""

    pass


class NXAPIPostError(NXAPIError):
    """Exception occurred during HTTP POST to NX-API."""

    pass


class NXAPIXMLError(NXAPIError):
    """Exception occurred processing XML response from NX-API."""

    pass



# -*- coding: utf-8 -*-
"""
NAPALM CLI Tools: configure
===========================

Deploy device config from the shell.
"""

# import helpers
from napalm.base import get_network_driver
from napalm.base.clitools.helpers import build_help
from napalm.base.clitools.helpers import configure_logging
from napalm.base.clitools.helpers import parse_optional_args
from napalm.base.clitools.helpers import warning

import sys
import logging

logger = logging.getLogger("cl-napalm-config.py")


def run(
    vendor, hostname, user, password, strategy, optional_args, config_file, dry_run
):

    logger.debug('Getting driver for OS "{driver}"'.format(driver=vendor))
    driver = get_network_driver(vendor)

    optional_args = parse_optional_args(optional_args)
    logger.debug(
        'Connecting to device "{}" with user "{}" and optional_args={}'.format(
            hostname, user, optional_args
        )
    )
    with driver(hostname, user, password, optional_args=optional_args) as device:
        logger.debug(
            'Strategy for loading configuration is "{strategy}"'.format(
                strategy=strategy
            )
        )
        if strategy == "replace":
            strategy_method = device.load_replace_candidate
        elif strategy == "merge":
            strategy_method = device.load_merge_candidate

        logger.debug('Loading configuration file "{config}"'.format(config=config_file))
        strategy_method(filename=config_file)

        logger.debug("Comparing configuration")
        diff = device.compare_config()

        if dry_run:
            logger.debug("Dry-run. Discarding configuration.")
        else:
            logger.debug("Committing configuration")
            device.commit_config()
        logger.debug("Closing session")
        return diff


def main():
    warning()
    args = build_help(configure=True)
    configure_logging(logger, args.debug)

    print(
        run(
            args.vendor,
            args.hostname,
            args.user,
            args.password,
            args.strategy,
            args.optional_args,
            args.config_file,
            args.dry_run,
        )
    )
    sys.exit(0)


if __name__ == "__main__":
    main()



"""Some functions to work with EOS version numbers"""
import re


class EOSVersion:
    """
    Class to represent EOS version
    """

    def __init__(self, version):
        """
        Create object
        :param version: str: version string
        """
        self.version = version
        self.numbers = []
        self.type = ""

        self._parse(version)

    def _parse(self, version):
        """
        Parse version string
        :param version: str: version
        :return: None
        """
        m = re.match(r"^(?P<numbers>\d[\d.]+\d)", version)

        if m:
            self.numbers = m.group("numbers").split(".")

    def __lt__(self, other):
        if not len(self.numbers):
            return True

        for x, y in zip(self.numbers, other.numbers):
            if x < y:
                return True
            elif x > y:
                return False

        return False

    def __gt__(self, other):
        if not len(self.numbers):
            return False

        for x, y in zip(self.numbers, other.numbers):
            if x > y:
                return True
            elif x < y:
                return False

        return False

    def __eq__(self, other):
        if len(self.numbers) != len(other.numbers):
            return False

        for x, y in zip(self.numbers, other.numbers):
            if x != y:
                return False

        return True

    def __le__(self, other):
        return self < other or self == other

    def __ge__(self, other):
        return self > other or self == other



"""
Validation methods for the NAPALM base.

See: https://napalm.readthedocs.io/en/latest/validate.html
"""
import yaml
import copy
import re

from napalm.base.exceptions import ValidationException


# We put it here to compile it only once
numeric_compare_regex = re.compile(r"^(<|>|<=|>=|==|!=)(\d+(\.\d+){0,1})$")


def _get_validation_file(validation_file):
    try:
        with open(validation_file, "r") as stream:
            try:
                validation_source = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                raise ValidationException(exc)
    except IOError:
        raise ValidationException("File {0} not found.".format(validation_file))
    return validation_source


def _mode(mode_string):
    mode = {"strict": False}

    for m in mode_string.split():
        if m not in mode.keys():
            raise ValidationException("mode '{}' not recognized".format(m))
        mode[m] = True
    return mode


def _compare_getter_list(src, dst, mode):
    result = {"complies": True, "present": [], "missing": [], "extra": []}
    for src_element in src:
        found = False

        i = 0
        while True:
            try:
                intermediate_match = compare(src_element, dst[i])
                if (
                    isinstance(intermediate_match, dict)
                    and intermediate_match["complies"]
                    or not isinstance(intermediate_match, dict)
                    and intermediate_match
                ):
                    found = True
                    result["present"].append(src_element)
                    dst.pop(i)
                    break
                else:
                    i += 1
            except IndexError:
                break

        if not found:
            result["complies"] = False
            result["missing"].append(src_element)

    if mode["strict"] and dst:
        result["extra"] = dst
        result["complies"] = False

    return result


def _compare_getter_dict(src, dst, mode):
    result = {"complies": True, "present": {}, "missing": [], "extra": []}
    dst = copy.deepcopy(dst)  # Otherwise we are going to modify a "live" object

    for key, src_element in src.items():
        try:
            dst_element = dst.pop(key)
            result["present"][key] = {}
            intermediate_result = compare(src_element, dst_element)

            if isinstance(intermediate_result, dict):
                nested = True

                complies = intermediate_result["complies"]

                if not complies:
                    result["present"][key]["diff"] = intermediate_result
            else:
                complies = intermediate_result
                nested = False
                if not complies:
                    result["present"][key]["expected_value"] = src_element
                    result["present"][key]["actual_value"] = dst_element

            if not complies:
                result["complies"] = False

            result["present"][key]["complies"] = complies
            result["present"][key]["nested"] = nested
        except KeyError:
            result["missing"].append(key)
            result["complies"] = False

    if mode["strict"] and dst:
        result["extra"] = list(dst.keys())
        result["complies"] = False

    return result


def compare(src, dst):
    if isinstance(src, str):
        src = str(src)

    if isinstance(src, dict):
        mode = _mode(src.pop("_mode", ""))
        if "list" in src.keys():
            if not isinstance(dst, list):
                # This can happen with nested lists
                return False

            return _compare_getter_list(src["list"], dst, mode)
        return _compare_getter_dict(src, dst, mode)

    elif isinstance(src, str):
        if src.startswith("<") or src.startswith(">"):
            cmp_result = _compare_numeric(src, dst)
            return cmp_result
        elif "<->" in src and len(src.split("<->")) == 2:
            cmp_result = _compare_range(src, dst)
            return cmp_result
        else:
            m = re.search(src, str(dst))
            if m:
                return bool(m)
            else:
                return src == dst

    elif type(src) == type(dst) == list:
        pairs = zip(src, dst)
        diff_lists = [
            [(k, x[k], y[k]) for k in x if not re.search(x[k], y[k])]
            for x, y in pairs
            if x != y
        ]
        return empty_tree(diff_lists)

    else:
        return src == dst


def _compare_numeric(src_num, dst_num):
    """Compare numerical values. You can use '<%d','>%d'."""
    dst_num = float(dst_num)

    match = numeric_compare_regex.match(src_num)
    if not match:
        error = "Failed numeric comparison. Collected: {}. Expected: {}".format(
            dst_num, src_num
        )
        raise ValueError(error)

    operand = {
        "<": "__lt__",
        ">": "__gt__",
        ">=": "__ge__",
        "<=": "__le__",
        "==": "__eq__",
        "!=": "__ne__",
    }
    return getattr(dst_num, operand[match.group(1)])(float(match.group(2)))


def _compare_range(src_num, dst_num):
    """Compare value against a range of values. You can use '%d<->%d'."""
    dst_num = float(dst_num)

    match = src_num.split("<->")
    if len(match) != 2:
        error = "Failed range comparison. Collected: {}. Expected: {}".format(
            dst_num, src_num
        )
        raise ValueError(error)

    if float(match[0]) <= dst_num <= float(match[1]):
        return True
    else:
        return False


def empty_tree(input_list):
    """Recursively iterate through values in nested lists."""
    for item in input_list:
        if not isinstance(item, list) or not empty_tree(item):
            return False
    return True


def compliance_report(cls, validation_file=None, validation_source=None):
    report = {}
    if validation_file:
        validation_source = _get_validation_file(validation_file)

    # Otherwise we are going to modify a "live" object
    validation_source = copy.deepcopy(validation_source)

    for validation_check in validation_source:
        for getter, expected_results in validation_check.items():
            if getter == "get_config":
                # TBD
                pass
            else:
                key = expected_results.pop("_name", "") or getter

                try:
                    kwargs = expected_results.pop("_kwargs", {})
                    actual_results = getattr(cls, getter)(**kwargs)
                    report[key] = compare(expected_results, actual_results)
                except NotImplementedError:
                    report[key] = {"skipped": True, "reason": "NotImplemented"}

    complies = all([e.get("complies", True) for e in report.values()])
    report["skipped"] = [k for k, v in report.items() if v.get("skipped", False)]
    report["complies"] = complies
    return report



"""Several methods to help with the tests."""


def test_model(model, data):
    """Return if the dictionary `data` complies with the `model`."""
    same_keys = set(model.keys()) == set(data.keys())

    if not same_keys:
        print(
            "model_keys: {}\ndata_keys: {}".format(
                sorted(model.keys()), sorted(data.keys())
            )
        )

    correct_class = True
    for key, instance_class in model.items():
        correct_class = isinstance(data[key], instance_class) and correct_class
        if not correct_class:
            print(
                "key: {}\nmodel_class: {}\ndata_class: {}".format(
                    key, instance_class, data[key].__class__
                )
            )

    return correct_class and same_keys



alive = {"is_alive": bool}

facts = {
    "os_version": str,
    "uptime": int,
    "interface_list": list,
    "vendor": str,
    "serial_number": str,
    "model": str,
    "hostname": str,
    "fqdn": str,
}

interface = {
    "is_up": bool,
    "is_enabled": bool,
    "description": str,
    "last_flapped": float,
    "mtu": int,
    "speed": int,
    "mac_address": str,
}

lldp_neighbors = {"hostname": str, "port": str}

interface_counters = {
    "tx_errors": int,
    "rx_errors": int,
    "tx_discards": int,
    "rx_discards": int,
    "tx_octets": int,
    "rx_octets": int,
    "tx_unicast_packets": int,
    "rx_unicast_packets": int,
    "tx_multicast_packets": int,
    "rx_multicast_packets": int,
    "tx_broadcast_packets": int,
    "rx_broadcast_packets": int,
}

temperature = {"is_alert": bool, "is_critical": bool, "temperature": float}

power = {"status": bool, "output": float, "capacity": float}

memory = {"used_ram": int, "available_ram": int}

fan = {"status": bool}

cpu = {"%usage": float}

peer = {
    "is_enabled": bool,
    "uptime": int,
    "remote_as": int,
    "description": str,
    "remote_id": str,
    "local_as": int,
    "is_up": bool,
    "address_family": dict,
}

af = {"sent_prefixes": int, "accepted_prefixes": int, "received_prefixes": int}

lldp_neighbors_detail = {
    "parent_interface": str,
    "remote_port": str,
    "remote_chassis_id": str,
    "remote_port_description": str,
    "remote_system_name": str,
    "remote_system_description": str,
    "remote_system_capab": list,
    "remote_system_enable_capab": list,
}

bgp_config_group = {
    "type": str,
    "description": str,
    "apply_groups": list,
    "multihop_ttl": int,
    "multipath": bool,
    "local_address": str,
    "local_as": int,
    "remote_as": int,
    "import_policy": str,
    "export_policy": str,
    "remove_private_as": bool,
    "prefix_limit": dict,
    "neighbors": dict,
}

bgp_config_neighbor = {
    "description": str,
    "import_policy": str,
    "export_policy": str,
    "local_address": str,
    "authentication_key": str,
    "nhs": bool,
    "route_reflector_client": bool,
    "local_as": int,
    "remote_as": int,
    "prefix_limit": dict,
}

peer_details = {
    "up": bool,
    "local_as": int,
    "remote_as": int,
    "router_id": str,
    "local_address": str,
    "routing_table": str,
    "local_address_configured": bool,
    "local_port": int,
    "remote_address": str,
    "remote_port": int,
    "multihop": bool,
    "multipath": bool,
    "remove_private_as": bool,
    "import_policy": str,
    "export_policy": str,
    "input_messages": int,
    "output_messages": int,
    "input_updates": int,
    "output_updates": int,
    "messages_queued_out": int,
    "connection_state": str,
    "previous_connection_state": str,
    "last_event": str,
    "suppress_4byte_as": bool,
    "local_as_prepend": bool,
    "holdtime": int,
    "configured_holdtime": int,
    "keepalive": int,
    "configured_keepalive": int,
    "active_prefix_count": int,
    "received_prefix_count": int,
    "accepted_prefix_count": int,
    "suppressed_prefix_count": int,
    "advertised_prefix_count": int,
    "flap_count": int,
}

arp_table = {"interface": str, "mac": str, "ip": str, "age": float}

ipv6_neighbor = {"interface": str, "mac": str, "ip": str, "age": float, "state": str}

ntp_peer = {
    # will populate it in the future wit potential keys
}

ntp_server = {
    # will populate it in the future wit potential keys
}

ntp_stats = {
    "remote": str,
    "referenceid": str,
    "synchronized": bool,
    "stratum": int,
    "type": str,
    "when": str,
    "hostpoll": int,
    "reachability": int,
    "delay": float,
    "offset": float,
    "jitter": float,
}

interfaces_ip = {"prefix_length": int}

mac_address_table = {
    "mac": str,
    "interface": str,
    "vlan": int,
    "static": bool,
    "active": bool,
    "moves": int,
    "last_move": float,
}

route = {
    "protocol": str,
    "current_active": bool,
    "last_active": bool,
    "age": int,
    "next_hop": str,
    "outgoing_interface": str,
    "selected_next_hop": bool,
    "preference": int,
    "inactive_reason": str,
    "routing_table": str,
    "protocol_attributes": dict,
}

snmp = {"chassis_id": str, "community": dict, "contact": str, "location": str}

snmp_community = {"acl": str, "mode": str}

probe_test = {
    "probe_type": str,
    "target": str,
    "source": str,
    "probe_count": int,
    "test_interval": int,
}

probe_test_results = {
    "target": str,
    "source": str,
    "probe_type": str,
    "probe_count": int,
    "rtt": float,
    "round_trip_jitter": float,
    "last_test_loss": int,
    "current_test_min_delay": float,
    "current_test_max_delay": float,
    "current_test_avg_delay": float,
    "last_test_min_delay": float,
    "last_test_max_delay": float,
    "last_test_avg_delay": float,
    "global_test_min_delay": float,
    "global_test_max_delay": float,
    "global_test_avg_delay": float,
}

ping = {
    "probes_sent": int,
    "packet_loss": int,
    "rtt_min": float,
    "rtt_max": float,
    "rtt_avg": float,
    "rtt_stddev": float,
    "results": list,
}

ping_result = {"ip_address": str, "rtt": float}

traceroute = {"rtt": float, "ip_address": str, "host_name": str}

users = {"level": int, "password": str, "sshkeys": list}

optics_state = {"instant": float, "avg": float, "min": float, "max": float}

config = {"running": str, "startup": str, "candidate": str}

network_instance = {"name": str, "type": str, "state": dict, "interfaces": dict}

network_instance_state = {"route_distinguisher": str}

network_instance_interfaces = {"interface": dict}

firewall_policies = {
    "position": int,
    "packet_hits": int,
    "byte_hits": int,
    "id": str,
    "enabled": bool,
    "schedule": str,
    "log": str,
    "l3_src": str,
    "l3_dst": str,
    "service": str,
    "src_zone": str,
    "dst_zone": str,
    "action": str,
}

vlan = {"name": str, "interfaces": list}



# Copyright 2015 Spotify AB. All rights reserved.
#
# The contents of this file are licensed under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
import difflib
from napalm.base import exceptions
from napalm.base.test import models
from unittest import SkipTest


class TestConfigNetworkDriver(object):
    @classmethod
    def setup_class(cls):
        """Added for py.test/nosetests compatibility"""
        cls.setUpClass()

    @classmethod
    def teardown_class(cls):
        """Added for py.test/nosetests compatibility"""
        cls.tearDownClass()

    @classmethod
    def tearDownClass(cls):
        cls.device.load_replace_candidate(filename="%s/initial.conf" % cls.vendor)
        cls.device.commit_config()
        cls.device.close()

    @staticmethod
    def read_file(filename):
        with open(filename, "r") as f:
            return f.read().strip()

    @staticmethod
    def print_diff_strings(orig, new):
        for line in difflib.context_diff(orig.splitlines(), new.splitlines()):
            print(line)

    def test_replacing_and_committing_config(self):
        try:
            self.device.load_replace_candidate(
                filename="%s/new_good.conf" % self.vendor
            )
            self.device.commit_config()
        except NotImplementedError:
            raise SkipTest()

        # The diff should be empty as the configuration has been committed already
        diff = self.device.compare_config()

        # Reverting changes
        self.device.load_replace_candidate(filename="%s/initial.conf" % self.vendor)
        self.device.commit_config()

        self.assertEqual(len(diff), 0)

    def test_replacing_config_with_typo(self):
        result = False
        try:
            self.device.load_replace_candidate(
                filename="%s/new_typo.conf" % self.vendor
            )
            self.device.commit_config()
        except NotImplementedError:
            raise SkipTest()
        except exceptions.ReplaceConfigException:
            self.device.load_replace_candidate(filename="%s/initial.conf" % self.vendor)
            diff = self.device.compare_config()
            self.device.discard_config()
            result = True and len(diff) == 0
        self.assertTrue(result)

    def test_replacing_config_and_diff_and_discard(self):
        intended_diff = self.read_file("%s/new_good.diff" % self.vendor)

        self.device.load_replace_candidate(filename="%s/new_good.conf" % self.vendor)
        commit_diff = self.device.compare_config()

        print(commit_diff)

        self.device.discard_config()
        discard_diff = self.device.compare_config()
        self.device.discard_config()

        result = (commit_diff == intended_diff) and (discard_diff == "")
        self.assertTrue(result)

    def test_replacing_config_and_rollback(self):
        self.device.load_replace_candidate(filename="%s/new_good.conf" % self.vendor)
        orig_diff = self.device.compare_config()
        self.device.commit_config()

        # Now we rollback changes
        replace_config_diff = self.device.compare_config()
        self.device.rollback()

        # We try to load the config again. If the rollback was successful new diff should be
        # like the first one
        self.device.load_replace_candidate(filename="%s/new_good.conf" % self.vendor)
        last_diff = self.device.compare_config()
        self.device.discard_config()

        result = (orig_diff == last_diff) and (len(replace_config_diff) == 0)

        self.assertTrue(result)

    def test_merge_configuration(self):
        intended_diff = self.read_file("%s/merge_good.diff" % self.vendor)

        self.device.load_merge_candidate(filename="%s/merge_good.conf" % self.vendor)
        self.device.commit_config()

        # Reverting changes
        self.device.load_replace_candidate(filename="%s/initial.conf" % self.vendor)
        diff = self.device.compare_config()

        print(diff)

        self.device.commit_config()

        self.assertEqual(diff, intended_diff)

    def test_merge_configuration_typo_and_rollback(self):
        result = False
        try:
            self.device.load_merge_candidate(
                filename="%s/merge_typo.conf" % self.vendor
            )
            self.device.compare_config()
            self.device.commit_config()
            raise Exception("We shouldn't be here")
        except exceptions.MergeConfigException:
            # We load the original config as candidate. If the commit failed cleanly the
            # compare_config should be empty
            self.device.load_replace_candidate(filename="%s/initial.conf" % self.vendor)
            result = self.device.compare_config() == ""
            self.device.discard_config()

        self.assertTrue(result)

    def test_load_template(self):
        """Test load_template method."""
        self.device.load_template("set_hostname", hostname="my-hostname")
        diff = self.device.compare_config()
        self.device.discard_config()
        self.assertTrue(diff != "")


class TestGettersNetworkDriver(object):
    @staticmethod
    def _test_model(model, data):
        same_keys = set(model.keys()) == set(data.keys())

        if not same_keys:
            print(
                "model_keys: {}\ndata_keys: {}".format(
                    sorted(model.keys()), sorted(data.keys())
                )
            )

        correct_class = True
        for key, instance_class in model.items():
            same_class = isinstance(data[key], instance_class)
            correct_class = correct_class and same_class
            if not same_class:
                print(
                    "key: {}\nmodel_class: {}\ndata_class: {}".format(
                        key, instance_class, data[key].__class__
                    )
                )

        return correct_class and same_keys

    def test_get_firewall_policies(self):
        try:
            policies = self.device.get_firewall_policies()
        except NotImplementedError:
            raise SkipTest()

        result = len(policies) > 0

        for policy_name, policy_details in policies.items():
            for policy_term in policy_details:
                result = result and self._test_model(
                    models.firewall_policies, policy_term
                )

        self.assertTrue(result)

    def test_is_alive(self):
        try:
            alive = self.device.is_alive()
        except NotImplementedError:
            raise SkipTest()
        result = self._test_model(models.alive, alive)
        self.assertTrue(result)

    def test_get_facts(self):
        try:
            facts = self.device.get_facts()
        except NotImplementedError:
            raise SkipTest()
        result = self._test_model(models.facts, facts)
        self.assertTrue(result)

    def test_get_interfaces(self):
        try:
            get_interfaces = self.device.get_interfaces()
        except NotImplementedError:
            raise SkipTest()
        result = len(get_interfaces) > 0

        for interface, interface_data in get_interfaces.items():
            result = result and self._test_model(models.interface, interface_data)

        self.assertTrue(result)

    def test_get_lldp_neighbors(self):
        try:
            get_lldp_neighbors = self.device.get_lldp_neighbors()
        except NotImplementedError:
            raise SkipTest()
        result = len(get_lldp_neighbors) > 0

        for interface, neighbor_list in get_lldp_neighbors.items():
            for neighbor in neighbor_list:
                result = result and self._test_model(models.lldp_neighbors, neighbor)

        self.assertTrue(result)

    def test_get_interfaces_counters(self):
        try:
            get_interfaces_counters = self.device.get_interfaces_counters()
        except NotImplementedError:
            raise SkipTest()
        result = len(self.device.get_interfaces_counters()) > 0

        for interface, interface_data in get_interfaces_counters.items():
            result = result and self._test_model(
                models.interface_counters, interface_data
            )

        self.assertTrue(result)

    def test_get_environment(self):
        try:
            environment = self.device.get_environment()
        except NotImplementedError:
            raise SkipTest()
        result = len(environment) > 0

        for fan, fan_data in environment["fans"].items():
            result = result and self._test_model(models.fan, fan_data)

        for power, power_data in environment["power"].items():
            result = result and self._test_model(models.power, power_data)

        for temperature, temperature_data in environment["temperature"].items():
            result = result and self._test_model(models.temperature, temperature_data)

        for cpu, cpu_data in environment["cpu"].items():
            result = result and self._test_model(models.cpu, cpu_data)

        result = result and self._test_model(models.memory, environment["memory"])

        self.assertTrue(result)

    def test_get_bgp_neighbors(self):
        try:
            get_bgp_neighbors = self.device.get_bgp_neighbors()
        except NotImplementedError:
            raise SkipTest()
        result = "global" in get_bgp_neighbors.keys()

        if not result:
            print("global is not part of the returned vrfs")
        else:
            for vrf, vrf_data in get_bgp_neighbors.items():
                result = result and isinstance(vrf_data["router_id"], str)
                if not result:
                    print("router_id is not {}".format(str))

                for peer, peer_data in vrf_data["peers"].items():
                    result = result and self._test_model(models.peer, peer_data)

                    for af, af_data in peer_data["address_family"].items():
                        result = result and self._test_model(models.af, af_data)

            self.assertTrue(result)

    def test_get_lldp_neighbors_detail(self):
        try:
            get_lldp_neighbors_detail = self.device.get_lldp_neighbors_detail()
        except NotImplementedError:
            raise SkipTest()
        result = len(get_lldp_neighbors_detail) > 0

        for interface, neighbor_list in get_lldp_neighbors_detail.items():
            for neighbor in neighbor_list:
                result = result and self._test_model(
                    models.lldp_neighbors_detail, neighbor
                )

        self.assertTrue(result)

    def test_get_bgp_config(self):
        try:
            get_bgp_config = self.device.get_bgp_config()
        except NotImplementedError:
            raise SkipTest()
        result = len(get_bgp_config) > 0

        for bgp_group in get_bgp_config.values():
            result = result and self._test_model(models.bgp_config_group, bgp_group)
            for bgp_neighbor in bgp_group.get("neighbors", {}).values():
                result = result and self._test_model(
                    models.bgp_config_neighbor, bgp_neighbor
                )

        self.assertTrue(result)

    def test_get_bgp_neighbors_detail(self):
        try:
            get_bgp_neighbors_detail = self.device.get_bgp_neighbors_detail()
        except NotImplementedError:
            raise SkipTest()

        result = len(get_bgp_neighbors_detail) > 0

        for vrf, vrf_ases in get_bgp_neighbors_detail.items():
            result = result and isinstance(vrf, str)
            for remote_as, neighbor_list in vrf_ases.items():
                result = result and isinstance(remote_as, int)
                for neighbor in neighbor_list:
                    result = result and self._test_model(models.peer_details, neighbor)

        self.assertTrue(result)

    def test_get_arp_table(self):
        try:
            get_arp_table = self.device.get_arp_table()
        except NotImplementedError:
            raise SkipTest()
        result = len(get_arp_table) > 0

        for arp_entry in get_arp_table:
            result = result and self._test_model(models.arp_table, arp_entry)

        self.assertTrue(result)

    def test_get_arp_table_with_vrf(self):
        try:
            get_arp_table = self.device.get_arp_table(vrf="TEST")
        except NotImplementedError:
            raise SkipTest()
        result = len(get_arp_table) > 0

        for arp_entry in get_arp_table:
            result = result and self._test_model(models.arp_table, arp_entry)

        self.assertTrue(result)

    def test_get_ntp_peers(self):
        try:
            get_ntp_peers = self.device.get_ntp_peers()
        except NotImplementedError:
            raise SkipTest()
        result = len(get_ntp_peers) > 0

        for peer, peer_details in get_ntp_peers.items():
            result = result and isinstance(peer, str)
            result = result and self._test_model(models.ntp_peer, peer_details)

        self.assertTrue(result)

    def test_get_ntp_servers(self):
        try:
            get_ntp_servers = self.device.get_ntp_servers()
        except NotImplementedError:
            raise SkipTest()

        result = len(get_ntp_servers) > 0

        for server, server_details in get_ntp_servers.items():
            result = result and isinstance(server, str)
            result = result and self._test_model(models.ntp_server, server_details)

        self.assertTrue(result)

    def test_get_ntp_stats(self):
        try:
            get_ntp_stats = self.device.get_ntp_stats()
        except NotImplementedError:
            raise SkipTest()
        result = len(get_ntp_stats) > 0

        for ntp_peer_details in get_ntp_stats:
            result = result and self._test_model(models.ntp_stats, ntp_peer_details)

        self.assertTrue(result)

    def test_get_interfaces_ip(self):
        try:
            get_interfaces_ip = self.device.get_interfaces_ip()
        except NotImplementedError:
            raise SkipTest()
        result = len(get_interfaces_ip) > 0

        for interface, interface_details in get_interfaces_ip.items():
            ipv4 = interface_details.get("ipv4", {})
            ipv6 = interface_details.get("ipv6", {})
            for ip, ip_details in ipv4.items():
                result = result and self._test_model(models.interfaces_ip, ip_details)
            for ip, ip_details in ipv6.items():
                result = result and self._test_model(models.interfaces_ip, ip_details)

        self.assertTrue(result)

    def test_get_mac_address_table(self):
        try:
            get_mac_address_table = self.device.get_mac_address_table()
        except NotImplementedError:
            raise SkipTest()
        result = len(get_mac_address_table) > 0

        for mac_table_entry in get_mac_address_table:
            result = result and self._test_model(
                models.mac_address_table, mac_table_entry
            )

        self.assertTrue(result)

    def test_get_route_to(self):
        destination = "1.0.4.0/24"
        protocol = "bgp"
        try:
            get_route_to = self.device.get_route_to(
                destination=destination, protocol=protocol
            )
        except NotImplementedError:
            raise SkipTest()

        result = len(get_route_to) > 0

        for prefix, routes in get_route_to.items():
            for route in routes:
                result = result and self._test_model(models.route, route)
        self.assertTrue(result)

    def test_get_snmp_information(self):
        try:
            get_snmp_information = self.device.get_snmp_information()
        except NotImplementedError:
            raise SkipTest()

        result = len(get_snmp_information) > 0

        for snmp_entry in get_snmp_information:
            result = result and self._test_model(models.snmp, get_snmp_information)

        for community, community_data in get_snmp_information["community"].items():
            result = result and self._test_model(models.snmp_community, community_data)

        self.assertTrue(result)

    def test_get_probes_config(self):
        try:
            get_probes_config = self.device.get_probes_config()
        except NotImplementedError:
            raise SkipTest()

        result = len(get_probes_config) > 0

        for probe_name, probe_tests in get_probes_config.items():
            for test_name, test_config in probe_tests.items():
                result = result and self._test_model(models.probe_test, test_config)

        self.assertTrue(result)

    def test_get_probes_results(self):
        try:
            get_probes_results = self.device.get_probes_results()
        except NotImplementedError:
            raise SkipTest()
        result = len(get_probes_results) > 0

        for probe_name, probe_tests in get_probes_results.items():
            for test_name, test_results in probe_tests.items():
                result = result and self._test_model(
                    models.probe_test_results, test_results
                )

        self.assertTrue(result)

    def test_ping(self):
        destination = "8.8.8.8"
        try:
            get_ping = self.device.ping(destination)
        except NotImplementedError:
            raise SkipTest()
        result = isinstance(get_ping.get("success"), dict)
        ping_results = get_ping.get("success", {})

        result = result and self._test_model(models.ping, ping_results)

        for ping_result in ping_results.get("results", []):
            result = result and self._test_model(models.ping_result, ping_result)

        self.assertTrue(result)

    def test_traceroute(self):

        destination = "8.8.8.8"
        try:
            get_traceroute = self.device.traceroute(destination)
        except NotImplementedError:
            raise SkipTest()
        result = isinstance(get_traceroute.get("success"), dict)
        traceroute_results = get_traceroute.get("success", {})

        for hope_id, hop_result in traceroute_results.items():
            for probe_id, probe_result in hop_result.get("probes", {}).items():
                result = result and self._test_model(models.traceroute, probe_result)

        self.assertTrue(result)

    def test_get_users(self):

        try:
            get_users = self.device.get_users()
        except NotImplementedError:
            raise SkipTest()
        result = len(get_users)

        for user, user_details in get_users.items():
            result = result and self._test_model(models.users, user_details)
            result = result and (0 <= user_details.get("level") <= 15)

        self.assertTrue(result)

    def test_get_optics(self):

        try:
            get_optics = self.device.get_optics()
        except NotImplementedError:
            raise SkipTest()

        assert isinstance(get_optics, dict)

        for iface, iface_data in get_optics.items():
            assert isinstance(iface, str)
            for channel in iface_data["physical_channels"]["channel"]:
                assert len(channel) == 2
                assert isinstance(channel["index"], int)
                for field in ["input_power", "output_power", "laser_bias_current"]:

                    assert len(channel["state"][field]) == 4
                    assert isinstance(channel["state"][field]["instant"], float)
                    assert isinstance(channel["state"][field]["avg"], float)
                    assert isinstance(channel["state"][field]["min"], float)
                    assert isinstance(channel["state"][field]["max"], float)

    def test_get_config(self):
        """Test get_config method."""
        try:
            get_config = self.device.get_config()
        except NotImplementedError:
            raise SkipTest()

        assert isinstance(get_config, dict)
        assert self._test_model(models.config, get_config)

    def test_get_config_filtered(self):
        """Test get_config method."""
        for config in ["running", "startup", "candidate"]:
            try:
                get_config = self.device.get_config(retrieve=config)
            except NotImplementedError:
                raise SkipTest()

            assert get_config["candidate"] == "" if config != "candidate" else True
            assert get_config["startup"] == "" if config != "startup" else True
            assert get_config["running"] == "" if config != "running" else True

    def test_get_network_instances(self):
        """Test get_network_instances method."""
        try:
            get_network_instances = self.device.get_network_instances()
        except NotImplementedError:
            raise SkipTest()

        result = isinstance(get_network_instances, dict)
        for network_instance_name, network_instance in get_network_instances.items():
            result = result and self._test_model(
                models.network_instance, network_instance
            )
            result = result and self._test_model(
                models.network_instance_state, network_instance["state"]
            )
            result = result and self._test_model(
                models.network_instance_interfaces, network_instance["interfaces"]
            )

        self.assertTrue(result)

    def test_get_ipv6_neighbors_table(self):
        try:
            get_ipv6_neighbors_table = self.device.get_ipv6_neighbors_table()
        except NotImplementedError:
            raise SkipTest()
        result = len(get_ipv6_neighbors_table) > 0

        for entry in get_ipv6_neighbors_table:
            result = result and self._test_model(models.ipv6_neighbor, entry)

        self.assertTrue(result)

    def test_get_vlans(self):
        try:
            get_vlans = self.device.get_vlans()
        except NotImplementedError:
            raise SkipTest()
        result = len(get_vlans) > 0

        for vlan, vlan_data in get_vlans.items():
            result = result and self._test_model(models.vlan, vlan_data)

        self.assertTrue(result)



"""Constants for the IOS-XR driver."""

from napalm.base.constants import *  # noqa

SR_638170159_SOLVED = False
# this flag says if the Cisco TAC SR 638170159
# has been solved
#
# "XML Agent Does not retrieve correct BGP routes data"
# is a weird bug reported on 2016-02-22 22:54:21
# briefly, all BGP routes are handled by the XML agent
# in such a way they have the following details:
#
# - all neighbors are 0.0.0.0
# - all routes are 0.0.0.0/0
# - all RD = 0000000000000000
#
# because of this none of the data retrieved
# from the BGP oper is usable thus has direct implications
# in our implementation of `get_route_to` when retrieving
# the BGP protocol specific attributes.



"""Constants to be used across NAPALM drivers."""

CONFIG_LOCK = True  # must be changed soon!
TIMEOUT = 60  # seconds

INTERFACE_NULL_SPEED = -1

ACTION_TYPE_METHODS = ("ping", "traceroute")

BGP_NEIGHBOR_NULL_COUNTER = -1

SNMP_AUTHORIZATION_MODE_MAP = {"read-only": "ro", "read-write": "rw"}

ROUTE_COMMON_PROTOCOL_FIELDS = [
    "destination",
    "prefix_length",
    "protocol",
    "current_active",
    "last_active",
    "age",
    "next_hop",
    "outgoing_interface",
    "selected_next_hop",
    "preference",
    "inactive_reason",
    "routing_table",
]  # identifies the list of fileds common for all protocols

ROUTE_PROTOCOL_SPECIFIC_FIELDS = {
    "bgp": [
        "local_as",
        "remote_as",
        "as_path",
        "communities",
        "local_preference",
        "preference2",
        "remote_address",
        "metric",
        "metric2",
    ],
    "isis": ["level", "metric", "local_as"],
    "static": [],  # nothing specific to static routes
}

TRACEROUTE_TTL = 255
TRACEROUTE_SOURCE = ""
TRACEROUTE_TIMEOUT = 2
TRACEROUTE_NULL_HOST_NAME = "*"
TRACEROUTE_NULL_IP_ADDRESS = "*"
TRACEROUTE_VRF = ""

OPTICS_NULL_LEVEL = "-Inf"

PING_SOURCE = ""
PING_TTL = 255
PING_TIMEOUT = 2
PING_SIZE = 100
PING_COUNT = 5
PING_VRF = ""

NETMIKO_MAP = {
    "ios": "cisco_ios",
    "nxos": "cisco_nxos",
    "nxos_ssh": "cisco_nxos",
    "iosxr": "cisco_iosxr",
    "eos": "arista_eos",
    "junos": "juniper_eos",
}
LLDP_CAPAB_TRANFORM_TABLE = {
    "o": "other",
    "p": "repeater",
    "b": "bridge",
    "w": "wlan-access-point",
    "r": "router",
    "t": "telephone",
    "c": "docsis-cable-device",
    "s": "station",
}

CISCO_SANITIZE_FILTERS = {
    r"^(snmp-server community).*$": r"\1 <removed>",
    r"^(snmp-server host \S+( vrf \S+)?( version (1|2c|3))?)\s+\S+((\s+\S*)*)\s*$": r"\1 <removed> \5",  # noqa
    r"^(snmp-server user \S+( \S+)? auth md5) \S+ (priv) \S+ (localizedkey( engineID \S+)?)\s*$": r"\1 <removed> \3 <removed> \4\5",  # noqa
    r"^(username .+ (password|secret) \d) .+$": r"\1 <removed>",
    r"^(enable (password|secret)( level \d+)? \d) .+$": r"\1 <removed>",
    r"^(\s+(?:password|secret)) (?:\d )?\S+$": r"\1 <removed>",
    r"^(.*wpa-psk ascii \d) (\S+)$": r"\1 <removed>",
    r"^(.*key 7) (\d.+)$": r"\1 <removed>",
    r"^(tacacs-server (.+ )?key) .+$": r"\1 <removed>",
    r"^(crypto isakmp key) (\S+) (.*)$": r"\1 <removed> \3",
    r"^(\s+ip ospf message-digest-key \d+ md5) .+$": r"\1 <removed>",
    r"^(\s+ip ospf authentication-key) .+$": r"\1 <removed>",
    r"^(\s+neighbor \S+ password) .+$": r"\1 <removed>",
    r"^(\s+vrrp \d+ authentication text) .+$": r"\1 <removed>",
    r"^(\s+standby \d+ authentication) .{1,8}$": r"\1 <removed>",
    r"^(\s+standby \d+ authentication md5 key-string) .+?( timeout \d+)?$": r"\1 <removed> \2",
    r"^(\s+key-string) .+$": r"\1 <removed>",
    r"^((tacacs|radius) server [^\n]+\n(\s+[^\n]+\n)*\s+key) [^\n]+$": r"\1 <removed>",
    r"^(\s+ppp (chap|pap) password \d) .+$": r"\1 <removed>",
}



# Copyright 2017 Dravetech AB. All rights reserved.
#
# The contents of this file are licensed under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

from napalm.base.base import NetworkDriver
import napalm.base.exceptions

import inspect
import json
import os
import re


from pydoc import locate


def raise_exception(result):
    exc = locate(result["exception"])
    if exc:
        raise exc(*result.get("args", []), **result.get("kwargs", {}))
    else:
        raise TypeError("Couldn't resolve exception {}", result["exception"])


def is_mocked_method(method):
    mocked_methods = []
    if method.startswith("get_") or method in mocked_methods:
        return True
    return False


def mocked_method(path, name, count):
    parent_method = getattr(NetworkDriver, name)
    parent_method_args = inspect.getfullargspec(parent_method)
    modifier = 0 if "self" not in parent_method_args.args else 1

    def _mocked_method(*args, **kwargs):
        # Check len(args)
        if len(args) + len(kwargs) + modifier > len(parent_method_args.args):
            raise TypeError(
                "{}: expected at most {} arguments, got {}".format(
                    name, len(parent_method_args.args), len(args) + modifier
                )
            )

        # Check kwargs
        unexpected = [x for x in kwargs if x not in parent_method_args.args]
        if unexpected:
            raise TypeError(
                "{} got an unexpected keyword argument '{}'".format(name, unexpected[0])
            )
        return mocked_data(path, name, count)

    return _mocked_method


def mocked_data(path, name, count):
    filename = "{}.{}".format(os.path.join(path, name), count)
    try:
        with open(filename) as f:
            result = json.loads(f.read())
    except IOError:
        raise NotImplementedError("You can provide mocked data in {}".format(filename))

    if "exception" in result:
        raise_exception(result)
    else:
        return result


class MockDevice(object):
    def __init__(self, parent, profile):
        self.parent = parent
        self.profile = profile

    def run_commands(self, commands):
        """Mock for EOS"""
        return list(self.parent.cli(commands).values())[0]

    def show(self, command):
        """Mock for nxos"""
        return self.run_commands([command])


class MockDriver(NetworkDriver):
    def __init__(self, hostname, username, password, timeout=60, optional_args=None):
        """
        Supported optional_args:
            * path(str) - path to where the mocked files are located
            * profile(list) - List of profiles to assign
        """
        self.hostname = hostname
        self.username = username
        self.password = password
        self.path = optional_args.get("path", "")
        self.profile = optional_args.get("profile", [])
        self.fail_on_open = optional_args.get("fail_on_open", False)

        self.opened = False
        self.calls = {}
        self.device = MockDevice(self, self.profile)

        # None no action, True load_merge, False load_replace
        self.merge = None
        self.filename = None
        self.config = None

    def _count_calls(self, name):
        current_count = self.calls.get(name, 0)
        self.calls[name] = current_count + 1
        return self.calls[name]

    def _raise_if_closed(self):
        if not self.opened:
            raise napalm.base.exceptions.ConnectionClosedException("connection closed")

    def open(self):
        if self.fail_on_open:
            raise napalm.base.exceptions.ConnectionException("You told me to do this")
        self.opened = True

    def close(self):
        self.opened = False

    def is_alive(self):
        return {"is_alive": self.opened}

    def cli(self, commands):
        count = self._count_calls("cli")
        result = {}
        regexp = re.compile("[^a-zA-Z0-9]+")
        for i, c in enumerate(commands):
            sanitized = re.sub(regexp, "_", c)
            name = "cli.{}.{}".format(count, sanitized)
            filename = "{}.{}".format(os.path.join(self.path, name), i)
            with open(filename, "r") as f:
                result[c] = f.read()
        return result

    def load_merge_candidate(self, filename=None, config=None):
        count = self._count_calls("load_merge_candidate")
        self._raise_if_closed()
        self.merge = True
        self.filename = filename
        self.config = config
        mocked_data(self.path, "load_merge_candidate", count)

    def load_replace_candidate(self, filename=None, config=None):
        count = self._count_calls("load_replace_candidate")
        self._raise_if_closed()
        self.merge = False
        self.filename = filename
        self.config = config
        mocked_data(self.path, "load_replace_candidate", count)

    def compare_config(self, filename=None, config=None):
        count = self._count_calls("compare_config")
        self._raise_if_closed()
        return mocked_data(self.path, "compare_config", count)["diff"]

    def commit_config(self):
        count = self._count_calls("commit_config")
        self._raise_if_closed()
        self.merge = None
        self.filename = None
        self.config = None
        mocked_data(self.path, "commit_config", count)

    def discard_config(self):
        count = self._count_calls("commit_config")
        self._raise_if_closed()
        self.merge = None
        self.filename = None
        self.config = None
        mocked_data(self.path, "discard_config", count)

    def _rpc(self, get):
        """This one is only useful for junos."""
        return list(self.cli([get]).values())[0]

    def __getattribute__(self, name):
        if is_mocked_method(name):
            self._raise_if_closed()
            count = self._count_calls(name)
            return mocked_method(self.path, name, count)
        else:
            return object.__getattribute__(self, name)



"""Test fixtures."""
import ast
import json
import os

NAPALM_TEST_MOCK = ast.literal_eval(os.getenv("NAPALM_TEST_MOCK", default="1"))
NAPALM_HOSTNAME = os.getenv("NAPALM_HOSTNAME", default="127.0.0.1")
NAPALM_USERNAME = os.getenv("NAPALM_USERNAME", default="vagrant")
NAPALM_PASSWORD = os.getenv("NAPALM_PASSWORD", default="vagrant")
NAPALM_OPTIONAL_ARGS = json.loads(
    os.getenv("NAPALM_OPTIONAL_ARGS", default='{"port": 12443}')
)


def set_device_parameters(request):
    """Set up the class."""
    if NAPALM_TEST_MOCK:
        driver = request.cls.patched_driver
    else:
        driver = request.cls.driver

    request.cls.device = driver(
        NAPALM_HOSTNAME,
        NAPALM_USERNAME,
        NAPALM_PASSWORD,
        timeout=60,
        optional_args=NAPALM_OPTIONAL_ARGS,
    )
    request.cls.device.open()


def pytest_generate_tests(metafunc, basefile):
    """Generate test cases dynamically."""
    if metafunc.function.__dict__.get("build_test_cases", False):
        path = os.path.join(
            os.path.dirname(basefile), "mocked_data", metafunc.function.__name__
        )

        if os.path.exists(path):
            sub_folders = os.listdir(path)
        else:
            sub_folders = []

        test_cases = []
        for test_case in sub_folders:
            if os.path.isdir(os.path.join(path, test_case)):
                test_cases.append(test_case)

        if not test_cases:
            test_cases.append("no_test_case_found")

        metafunc.parametrize("test_case", test_cases)



"""Some common jinja filters."""


class CustomJinjaFilters(object):
    """Utility filters for jinja2."""

    @classmethod
    def filters(cls):
        """Return jinja2 filters that this module provide."""
        return {
            "oc_attr_isdefault": oc_attr_isdefault,
            "openconfig_to_cisco_af": openconfig_to_cisco_af,
            "openconfig_to_eos_af": openconfig_to_eos_af,
        }


def oc_attr_isdefault(o):
    """Return wether an OC attribute has been defined or not."""
    if not o._changed() and not o.default():
        return True
    if o == o.default():
        return True
    return False


def openconfig_to_cisco_af(value):
    """Translate openconfig AF name to Cisco AFI name."""
    if ":" in value:
        value = value.split(":")[1]

    mapd = {
        "IPV4_UNICAST": "ipv4 unicast",
        "IPV6_UNICAST": "ipv6 unicast",
        "IPV4_LABELED_UNICAST": "ipv4 unicast",
        "IPV6_LABELED_UNICAST": "ipv6 unicast",
        "L3VPN_IPV4_UNICAST": "vpnv4",
        "L3VPN_IPV6_UNICAST": "vpnv6",
    }
    return mapd[value]


def openconfig_to_eos_af(value):
    """Translate openconfig AF name to EOS AFI name."""
    if ":" in value:
        value = value.split(":")[1]

    mapd = {"IPV4_UNICAST": "ipv4", "IPV6_UNICAST": "ipv6"}
    return mapd[value]



# Copyright 2016 Dravetech AB. All rights reserved.
#
# The contents of this file are licensed under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

"""napalm.eos package."""
from napalm.eos.eos import EOSDriver

__all__ = ("EOSDriver",)



"""
Fork of pynxos library from network to code and mzbenami

Re-implemented by ktbyers to support XML-RPC in addition to JSON-RPC
"""
from napalm.nxapi_plumbing.device import Device
from napalm.nxapi_plumbing.api_client import RPCClient, XMLClient
from napalm.nxapi_plumbing.errors import (
    NXAPIError,
    NXAPICommandError,
    NXAPIConnectionError,
    NXAPIAuthError,
    NXAPIPostError,
    NXAPIXMLError,
)

__version__ = "0.6.0"
__all__ = (
    "Device",
    "RPCClient",
    "XMLClient",
    "NXAPIError",
    "NXAPICommandError",
    "NXAPIConnectionError",
    "NXAPIAuthError",
    "NXAPIPostError",
    "NXAPIXMLError",
)



import pkg_resources
import sys
import logging

from napalm.base import get_network_driver
from napalm._SUPPORTED_DRIVERS import SUPPORTED_DRIVERS

# Load constituent packages to fix threading deadlock on get_network_driver()
import napalm.eos
import napalm.ios
import napalm.iosxr
import napalm.junos
import napalm.nxos
import napalm.nxos_ssh
import napalm.pyIOSXR
import napalm.nxapi_plumbing  # noqa

# Verify Python Version that is running
try:
    if not (sys.version_info.major == 3 and sys.version_info.minor >= 6):
        raise RuntimeError("NAPALM requires Python 3.6 or greater")
except AttributeError:
    raise RuntimeError("NAPALM requires Python 3.6 or greater")

try:
    __version__ = pkg_resources.get_distribution("napalm").version
except pkg_resources.DistributionNotFound:
    __version__ = "Not installed"

__all__ = ("get_network_driver", "SUPPORTED_DRIVERS")

logger = logging.getLogger("napalm")



"""Constants for the JunOS driver."""

from napalm.base.constants import *  # noqa

# OpenConfig mapping
# ref: https://github.com/openconfig/public/blob/master/release/models/network-instance/openconfig-network-instance-types.yang  # noqa
OC_NETWORK_INSTANCE_TYPE_MAP = {
    "default": "DEFAULT_INSTANCE",
    "l2vpn": "L2VPN",
    "vrf": "L3VRF",
    "evpn": "BGP_EVPN",
    "vpls": "BGP_VPLS",
    "forwarding": "L2P2P",
}
# OPTICS_NULL_LEVEL_SPC matches infinite light level '- Inf'
# reading on some versions of JUNOS
# https://github.com/napalm-automation/napalm/issues/491
OPTICS_NULL_LEVEL_SPC = "- Inf"

LLDP_CAPAB_TRANFORM_TABLE = {
    "other": "other",
    "repeater": "repeater",
    "bridge": "bridge",
    "wlan access point": "wlan-access-point",
    "router": "router",
    "telephone": "telephone",
    "docsis cable device": "docsis-cable-device",
    "station only": "station",
}



"""Base class for Test doubles."""
import json
import re
import os
import sys


class BaseTestDouble(object):
    """Base class for test doubles."""

    def __init__(self, *args, **kwargs):
        """Initiate object."""
        self.current_test = ""
        self.current_test_case = ""

    def find_file(self, filename):
        """Find the necessary file for the given test case."""
        # Find base_dir of submodule
        module_dir = os.path.dirname(sys.modules[self.__module__].__file__)

        full_path = os.path.join(
            module_dir,
            "mocked_data",
            self.current_test,
            self.current_test_case,
            filename,
        )

        if os.path.exists(full_path):
            return full_path
        else:
            raise IOError("Couldn't find file with mocked data: {}".format(full_path))

    @staticmethod
    def sanitize_text(text):
        """Remove some weird characters from text, useful for building filenames from commands."""
        regexp = "[^a-zA-Z0-9]"
        return re.sub(regexp, "_", text)[0:150]

    @staticmethod
    def read_json_file(filename):
        """Parse a json file and return its content."""
        with open(filename) as data_file:
            return json.load(data_file)

    @staticmethod
    def read_txt_file(filename):
        """Return the content of a file."""
        with open(filename) as data_file:
            return data_file.read()

    @property
    def expected_result(self):
        """Return the expected result for the current test case."""
        filename = self.find_file("expected_result.json")

        with open(filename, mode="r") as f:
            try:
                return json.loads(f.read())
            except ValueError:
                raise ValueError(
                    "No JSON object could be decoded on filename: {}".format(filename)
                )


def _string_key_to_int(param):
    """For a given dictionary, convert all strings that represent a number into an int."""
    new_dict = {}

    if isinstance(param, list):
        return [_string_key_to_int(element) for element in param]
    elif isinstance(param, dict):
        for key, value in param.items():
            try:
                new_dict[int(key)] = _string_key_to_int(value)
            except ValueError:
                new_dict[key] = _string_key_to_int(value)
        return new_dict
    else:
        return param



"""
Load tables/views
"""
import yaml
import re
from jnpr.junos.factory import FactoryLoader
from os.path import splitext


def _preprocess_yml(path):
    """Dynamically create PY3 version of the file by re-writing 'unicode' to 'str'."""
    with open(path) as f:
        tmp_yaml = f.read()
    return re.sub(r"unicode", "str", tmp_yaml)


def _loadyaml_bypass(yaml_str):
    """Bypass Juniper's loadyaml and directly call FactoryLoader"""
    return FactoryLoader().load(yaml.safe_load(yaml_str))


_YAML_ = splitext(__file__)[0] + ".yml"
py3_yaml = _preprocess_yml(_YAML_)
globals().update(_loadyaml_bypass(py3_yaml))



"""
Fork of pynxos library from network to code and mzbenami

Reimplemented by ktbyers to support XML-RPC in addition to JSON-RPC
"""

from __future__ import print_function, unicode_literals

from napalm.nxapi_plumbing.errors import NXAPIError, NXAPICommandError
from napalm.nxapi_plumbing.api_client import RPCClient, XMLClient


class Device(object):
    def __init__(
        self,
        host,
        username,
        password,
        transport="http",
        api_format="jsonrpc",
        port=None,
        timeout=30,
        verify=True,
    ):
        self.host = host
        self.username = username
        self.password = password
        self.transport = transport
        self.api_format = api_format
        self.verify = verify
        self.port = port

        if api_format == "xml":
            self.api = XMLClient(
                host,
                username,
                password,
                transport=transport,
                port=port,
                timeout=timeout,
                verify=verify,
            )
        elif api_format == "jsonrpc":
            self.api = RPCClient(
                host,
                username,
                password,
                transport=transport,
                port=port,
                timeout=timeout,
                verify=verify,
            )

    def show(self, command, raw_text=False):
        """Send a non-configuration command.

        Args:
            command (str): The command to send to the device.

        Keyword Args:
            raw_text (bool): Whether to return raw text or structured data.

        Returns:
            The output of the show command, which could be raw text or structured data.
        """
        commands = [command]
        result = self.show_list(commands, raw_text)
        if len(result) > 1:
            raise NXAPIError(
                "Length of response inconsistent with number of commands executed."
            )

        # Return the only entry or the empty response
        if result:
            if self.api_format == "jsonrpc":
                return result[0]["result"]
            elif self.api_format == "xml":
                return result[0]

        return result

    def show_list(self, commands, raw_text=False):
        """Send a list of non-configuration commands.

        Args:
            commands (list): A list of commands to send to the device.

        Keyword Args:
            raw_text (bool): Whether to return raw text or structured data.

        Returns:
            A list of outputs for each show command
        """
        cmd_method = self.api.cmd_method_raw if raw_text else self.api.cmd_method
        return self.api._nxapi_command(commands, method=cmd_method)

    def config(self, command):
        """Send a configuration command.

        Args:
            command (str): The command to send to the device.

        Raises:
            NXAPICommandError: If there is a problem with the supplied command.
        """
        commands = [command]
        result = self.config_list(commands)

        if len(result) > 1:
            raise NXAPIError(
                "Length of response inconsistent with number of commands executed."
            )

        # Return the only entry or the empty response
        if result:
            if self.api_format == "jsonrpc":
                return result[0]["result"]
            elif self.api_format == "xml":
                return result[0]

        return result

    def config_list(self, commands):
        """Send a list of configuration commands.

        Args:
            commands (list): A list of commands to send to the device.

        Raises:
            NXAPICommandError: If there is a problem with one of the commands in the list.
        """
        return self.api._nxapi_command_conf(commands)

    def save(self, filename="startup-config"):
        """Save a device's running configuration.

        Args:
            filename (str): The filename on the remote device.
                If none is supplied, the implementing class should
                save to the "startup configuration".
        """
        try:
            cmd = "copy run {}".format(filename)
            self.show(cmd, raw_text=True)
        except NXAPICommandError as e:
            if "overwrite" in e.message:
                return False
            raise
        return True

    def rollback(self, filename):
        """Rollback to a checkpoint file.

        Args:
            filename (str): The filename of the checkpoint file to load into the running
            configuration.
        """
        cmd = "rollback running-config file {}".format(filename)
        self.show(cmd, raw_text=True)

    def checkpoint(self, filename):
        """Save a checkpoint of the running configuration to the device.

        Args:
            filename (str): The filename to save the checkpoint as on the remote device.
        """
        self.show_list(
            ["terminal dont-ask", "checkpoint file {}".format(filename)], raw_text=True
        )



# import helpers
from napalm.base import get_network_driver
from napalm.base.clitools import helpers

# stdlib
import json
import logging
import argparse
import getpass
import pkg_resources
from functools import wraps


def debugging(name):
    def real_decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            censor_parameters = ["password"]
            censored_kwargs = {
                k: v if k not in censor_parameters else "*******"
                for k, v in kwargs.items()
            }
            logger.debug(
                "{} - Calling with args: {}, {}".format(name, args, censored_kwargs)
            )
            try:
                r = func(*args, **kwargs)
                logger.debug("{} - Successful".format(name))
                return r
            except NotImplementedError:
                if name not in [
                    "pre_connection_tests",
                    "connection_tests",
                    "post_connection_tests",
                ]:
                    logger.debug("{} - Not implemented".format(name))
            except Exception as e:
                logger.error("{} - Failed: {}".format(name, e))
                print("\n================= Traceback =================\n")
                raise

        return wrapper

    return real_decorator


logger = logging.getLogger("napalm")


def build_help():
    parser = argparse.ArgumentParser(
        description="Command line tool to handle configuration on devices using NAPALM."
        "The script will print the diff on the screen",
        epilog="Automate all the things!!!",
    )
    parser.add_argument(
        dest="hostname",
        action="store",
        help="Host where you want to deploy the configuration.",
    )
    parser.add_argument(
        "--user",
        "-u",
        dest="user",
        action="store",
        default=getpass.getuser(),
        help="User for authenticating to the host. Default: user running the script.",
    )
    parser.add_argument(
        "--password",
        "-p",
        dest="password",
        action="store",
        help="Password for authenticating to the host."
        "If you do not provide a password in the CLI you will be prompted.",
    )
    parser.add_argument(
        "--vendor",
        "-v",
        dest="vendor",
        action="store",
        required=True,
        help="Host Operating System.",
    )
    parser.add_argument(
        "--optional_args",
        "-o",
        dest="optional_args",
        action="store",
        help="String with comma separated key=value pairs passed via optional_args to the driver.",
    )
    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        help="Enables debug mode; more verbosity.",
    )
    subparser = parser.add_subparsers(title="actions")

    config = subparser.add_parser("configure", help="Perform a configuration operation")
    config.set_defaults(which="config")
    config.add_argument(
        dest="config_file",
        action="store",
        help="File containing the configuration you want to deploy.",
    )
    config.add_argument(
        "--strategy",
        "-s",
        dest="strategy",
        action="store",
        choices=["replace", "merge"],
        default="replace",
        help="Strategy to use to deploy configuration. Default: replace.",
    )
    config.add_argument(
        "--dry-run",
        "-d",
        dest="dry_run",
        action="store_true",
        default=None,
        help="Only returns diff, it does not deploy the configuration.",
    )

    call = subparser.add_parser("call", help="Call a napalm method")
    call.set_defaults(which="call")
    call.add_argument(dest="method", action="store", help="Run this method")
    call.add_argument(
        "--method-kwargs",
        "-k",
        dest="method_kwargs",
        action="store",
        help='kwargs to pass to the method. For example: "destination=1.1.1.1,protocol=bgp"',
    )

    validate = subparser.add_parser("validate", help="Validate configuration/state")
    validate.set_defaults(which="validate")
    validate.add_argument(
        dest="validation_file",
        action="store",
        help="Validation file containing resources derised states",
    )
    args = parser.parse_args()

    if not hasattr(args, "which"):
        args.which = None

    if args.password is None:
        password = getpass.getpass("Enter password: ")
        setattr(args, "password", password)

    return args


def check_installed_packages():
    logger.debug("Gathering napalm packages")
    napalm_packages = sorted(
        [
            "{}=={}".format(i.key, i.version)
            for i in pkg_resources.working_set
            if i.key.startswith("napalm")
        ]
    )
    for n in napalm_packages:
        logger.debug(n)


@debugging("get_network_driver")
def call_get_network_driver(vendor):
    return get_network_driver(vendor)


@debugging("__init__")
def call_instantiating_object(driver, *args, **kwargs):
    return driver(*args, **kwargs)


@debugging("pre_connection_tests")
def call_pre_connection(driver):
    driver.pre_connection_tests()


@debugging("connection_tests")
def call_connection(device):
    device.connection_tests()


@debugging("post_connection_tests")
def call_post_connection(device):
    device.post_connection_tests()


@debugging("get_facts")
def call_facts(device):
    facts = device.get_facts()
    logger.debug("Gathered facts:\n{}".format(json.dumps(facts, indent=4)))
    print(json.dumps(facts, indent=4))


@debugging("close")
def call_close(device):
    return device.close()


@debugging("open")
def call_open_device(device):
    return device.open()


@debugging("load_replace_candidate")
def call_load_replace_candidate(device, *args, **kwargs):
    return device.load_replace_candidate(*args, **kwargs)


@debugging("load_merge_candidate")
def call_load_merge_candidate(device, *args, **kwargs):
    return device.load_merge_candidate(*args, **kwargs)


@debugging("compare_config")
def call_compare_config(device, *args, **kwargs):
    diff = device.compare_config(*args, **kwargs)
    logger.debug("Gathered diff:")
    print(diff)
    return diff


@debugging("commit_config")
def call_commit_config(device, *args, **kwargs):
    return device.commit_config(*args, **kwargs)


def configuration_change(device, config_file, strategy, dry_run):
    if strategy == "replace":
        strategy_method = call_load_replace_candidate
    elif strategy == "merge":
        strategy_method = call_load_merge_candidate

    strategy_method(device, filename=config_file)

    diff = call_compare_config(device)

    if not dry_run:
        call_commit_config(device)
    return diff


@debugging("method")
def call_getter(device, method, **kwargs):
    logger.debug("{} - Attempting to resolve method".format(method))
    func = getattr(device, method)
    logger.debug(
        "{} - Attempting to call method with kwargs: {}".format(method, kwargs)
    )
    r = func(**kwargs)
    logger.debug("{} - Response".format(method))
    print(json.dumps(r, indent=4))


@debugging("compliance_report")
def call_compliance_report(device, validation_file):
    result = device.compliance_report(validation_file)
    print(json.dumps(result, indent=4))
    return result


def run_tests(args):
    driver = call_get_network_driver(args.vendor)
    optional_args = helpers.parse_optional_args(args.optional_args)

    device = call_instantiating_object(
        driver,
        args.hostname,
        args.user,
        password=args.password,
        timeout=60,
        optional_args=optional_args,
    )

    if args.debug:
        call_pre_connection(device)

    call_open_device(device)

    if args.debug:
        call_connection(device)
        call_facts(device)

    if args.which == "call":
        method_kwargs = helpers.parse_optional_args(args.method_kwargs)
        call_getter(device, args.method, **method_kwargs)
    elif args.which == "config":
        configuration_change(device, args.config_file, args.strategy, args.dry_run)
    elif args.which == "validate":
        call_compliance_report(device, args.validation_file)

    call_close(device)

    if args.debug:
        call_post_connection(device)


def main():
    args = build_help()
    helpers.configure_logging(logger, debug=args.debug)
    logger.debug("Starting napalm's debugging tool")
    check_installed_packages()
    run_tests(args)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
NAPALM CLI Tools: validate
===========================

Validating deployments from the shell.
"""

# import helpers
from napalm.base import get_network_driver
from napalm.base.clitools.helpers import build_help
from napalm.base.clitools.helpers import configure_logging
from napalm.base.clitools.helpers import parse_optional_args
from napalm.base.clitools.helpers import warning

# stdlib
import sys
import json
import logging

logger = logging.getLogger("cl_napalm_validate.py")


def main():
    warning()
    args = build_help(validate=True)
    configure_logging(logger, args.debug)

    logger.debug('Getting driver for OS "{driver}"'.format(driver=args.vendor))
    driver = get_network_driver(args.vendor)

    optional_args = parse_optional_args(args.optional_args)
    logger.debug(
        'Connecting to device "{}" with user "{}" and optional_args={}'.format(
            args.hostname, args.user, optional_args
        )
    )

    with driver(
        args.hostname, args.user, args.password, optional_args=optional_args
    ) as device:
        logger.debug("Generating compliance report")
        print(json.dumps(device.compliance_report(args.validation_file), indent=4))
        logger.debug("Closing session")
    sys.exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# coding=utf-8
# Copyright 2015 Netflix. All rights reserved.
# Copyright 2016 BigWaveIT. All rights reserved.
#
# The contents of this file are licensed under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

"""
A module to interact with Cisco devices running IOS-XR.

Forked from https://github.com/fooelisa/pyiosxr 2019-10-27

Contributors fooelisa, mirceaulinic, ktbyers et all
"""

from napalm.pyIOSXR.iosxr import IOSXR

__all__ = ("IOSXR",)



# Copyright 2015 Spotify AB. All rights reserved.
#
# The contents of this file are licensed under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

"""napalm.base package."""

# Python std lib
import inspect
import importlib

# NAPALM base
from napalm.base.base import NetworkDriver
from napalm.base.exceptions import ModuleImportError
from napalm.base.mock import MockDriver

__all__ = [
    "get_network_driver",  # export the function
    "NetworkDriver",  # also export the base class
]


def get_network_driver(name, prepend=True):
    """
    Searches for a class derived form the base NAPALM class NetworkDriver in a specific library.
    The library name must repect the following pattern: napalm_[DEVICE_OS].
    NAPALM community supports a list of devices and provides the corresponding libraries; for
    full reference please refer to the `Supported Network Operation Systems`_ paragraph on
    `Read the Docs`_.

    .. _`Supported Network Operation Systems`: \
    http://napalm.readthedocs.io/en/latest/#supported-network-operating-systems
    .. _`Read the Docs`: \
    http://napalm.readthedocs.io/

    :param name:         the name of the device operating system or the name of the library.
    :return:                    the first class derived from NetworkDriver, found in the library.
    :raise ModuleImportError:   when the library is not installed or a derived class from \
    NetworkDriver was not found.

    Example::

    .. code-block:: python

        >>> get_network_driver('junos')
        <class 'napalm.junos.junos.JunOSDriver'>
        >>> get_network_driver('IOS-XR')
        <class 'napalm.iosxr.iosxr.IOSXRDriver'>
        >>> get_network_driver('napalm.eos')
        <class 'napalm.eos.eos.EOSDriver'>
        >>> get_network_driver('wrong')
        napalm.base.exceptions.ModuleImportError: Cannot import "napalm_wrong". Is the library \
        installed?
    """
    if name == "mock":
        return MockDriver

    if not (isinstance(name, str) and len(name) > 0):
        raise ModuleImportError("Please provide a valid driver name.")

    # Only lowercase allowed
    name = name.lower()
    # Try to not raise error when users requests IOS-XR for e.g.
    module_install_name = name.replace("-", "")
    community_install_name = "napalm_{name}".format(name=module_install_name)
    custom_install_name = "custom_napalm.{name}".format(name=module_install_name)
    # Can also request using napalm_[SOMETHING]
    if "napalm" not in module_install_name and prepend is True:
        module_install_name = "napalm.{name}".format(name=module_install_name)
    # Order is custom_napalm_os (local only) ->  napalm.os (core) ->  napalm_os (community)
    for module_name in [
        custom_install_name,
        module_install_name,
        community_install_name,
    ]:
        try:
            module = importlib.import_module(module_name)
            break
        except ImportError as e:
            message = str(e)
            if "No module named" in message:
                # py2 doesn't have ModuleNotFoundError exception
                failed_module = message.split()[-1]
                if failed_module.replace("'", "") in module_name:
                    continue
            raise e
    else:
        raise ModuleImportError(
            'Cannot import "{install_name}". Is the library installed?'.format(
                install_name=name
            )
        )

    for name, obj in inspect.getmembers(module):
        if inspect.isclass(obj) and issubclass(obj, NetworkDriver):
            return obj

    # looks like you don't have any Driver class in your module...
    raise ModuleImportError(
        'No class inheriting "napalm.base.base.NetworkDriver" found in "{install_name}".'.format(
            install_name=module_install_name
        )
    )



# Copyright 2015 Spotify AB. All rights reserved.
#
# The contents of this file are licensed under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.


class NapalmException(Exception):
    """
    Base Exception Class.
    """

    pass


class ModuleImportError(NapalmException):
    pass


class ConnectionException(NapalmException):
    """
    Unable to connect to the network device.
    """

    pass


class ConnectAuthError(ConnectionException):
    """
    Unable to connect to the network device
    due to invalid credentials.
    """

    pass


class ConnectTimeoutError(ConnectionException):
    """
    Exception raised when the connection to the
    network device takes too long.
    This may be avoided by adjusting the `timeout`
    argument.
    """

    pass


class ConnectionClosedException(ConnectionException):
    """
    The network device closed the connection.
    Raised whenever we try to execute a certain
    function, but we detect that the connection
    is not usable anymore. This can happen for
    various reasons: the network device terminates the
    session or it is dropped by a firewall or
    the server.
    """

    pass


class ReplaceConfigException(NapalmException):
    pass


class MergeConfigException(NapalmException):
    pass


class CommitError(NapalmException):
    """
    Raised when unable to commit the candidate config
    into the running config.
    """

    pass


class LockError(NapalmException):
    """
    Unable to lock the candidate config.
    """

    pass


class UnlockError(NapalmException):
    """
    Unable to unlock the candidate config.
    """

    pass


class SessionLockedException(NapalmException):
    pass


class CommandTimeoutException(NapalmException):
    pass


class CommandErrorException(NapalmException):
    pass


class DriverTemplateNotImplemented(NapalmException):
    pass


class TemplateNotImplemented(NapalmException):
    pass


class TemplateRenderException(NapalmException):
    pass


class ValidationException(NapalmException):
    pass



"""
Fork of pynxos library from network to code and mzbenami

Reimplemented by ktbyers to support XML-RPC in addition to JSON-RPC
"""
from __future__ import print_function, unicode_literals

from builtins import super
import requests
from requests.auth import HTTPBasicAuth
from requests.exceptions import ConnectionError
import json

from lxml import etree

from six import string_types

from napalm.nxapi_plumbing.errors import (
    NXAPIError,
    NXAPIPostError,
    NXAPICommandError,
    NXAPIXMLError,
    NXAPIAuthError,
    NXAPIConnectionError,
)


class RPCBase(object):
    """RPCBase class should be API-type neutral (i.e. shouldn't care whether XML or jsonrpc)."""

    def __init__(
        self,
        host,
        username,
        password,
        transport="https",
        port=None,
        timeout=30,
        verify=True,
    ):
        if transport not in ["http", "https"]:
            raise NXAPIError("'{}' is an invalid transport.".format(transport))

        if port is None:
            if transport == "http":
                port = 80
            elif transport == "https":
                port = 443

        self.url = "{}://{}:{}/ins".format(transport, host, port)
        self.username = username
        self.password = password
        self.timeout = timeout
        self.verify = verify

    def _process_api_response(self, response, commands, raw_text=False):
        raise NotImplementedError("Method must be implemented in child class")

    def _send_request(self, commands, method):
        payload = self._build_payload(commands, method)

        try:
            response = requests.post(
                self.url,
                timeout=self.timeout,
                data=payload,
                headers=self.headers,
                auth=HTTPBasicAuth(self.username, self.password),
                verify=self.verify,
            )
        except ConnectionError as e:
            raise NXAPIConnectionError(str(e))

        if response.status_code == 401:
            msg = (
                "Authentication to NX-API failed please verify your username, password, "
                "and hostname."
            )
            raise NXAPIAuthError(msg)

        if response.status_code not in [200]:
            msg = """Invalid status code returned on NX-API POST
commands: {}
status_code: {}""".format(
                commands, response.status_code
            )
            raise NXAPIPostError(msg)

        return response.text


class RPCClient(RPCBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.headers = {"content-type": "application/json-rpc"}
        self.api = "jsonrpc"
        self.cmd_method = "cli"
        self.cmd_method_conf = "cli"
        self.cmd_method_raw = "cli_ascii"

    def _nxapi_command(self, commands, method=None):
        """Send a command down the NX-API channel."""
        if method is None:
            method = self.cmd_method
        if isinstance(commands, string_types):
            commands = [commands]

        raw_text = True if method == "cli_ascii" else False

        response = self._send_request(commands, method=method)
        api_response = self._process_api_response(response, commands, raw_text=raw_text)
        return api_response

    def _nxapi_command_conf(self, commands, method=None):
        if method is None:
            method = self.cmd_method_conf
        return self._nxapi_command(commands=commands, method=method)

    def _build_payload(self, commands, method, rpc_version="2.0", api_version=1.0):
        """Construct the JSON-RPC payload for NX-API."""
        payload_list = []
        id_num = 1
        for command in commands:
            payload = {
                "jsonrpc": rpc_version,
                "method": method,
                "params": {"cmd": command, "version": api_version},
                "id": id_num,
            }
            payload_list.append(payload)
            id_num += 1

        return json.dumps(payload_list)

    def _process_api_response(self, response, commands, raw_text=False):
        """
        Normalize the API response including handling errors; adding the sent command into
        the returned data strucutre; make response structure consistent for raw_text and
        structured data.
        """

        response_list = json.loads(response)
        if isinstance(response_list, dict):
            response_list = [response_list]

        # Add the 'command' that was executed to the response dictionary
        for i, response_dict in enumerate(response_list):
            response_dict["command"] = commands[i]

        new_response = []
        for response in response_list:

            # Dectect errors
            self._error_check(response)

            # Some commands like "show run" can have a None result
            cmd_response = response.get("result")
            if cmd_response is None:
                cmd_response = {}

            # Normalize the response data structure
            response_dict = {"command": response["command"]}
            if response and raw_text:
                response_dict["result"] = cmd_response.get("msg")
            elif response and not raw_text:
                response_dict["result"] = cmd_response.get("body")
            else:
                raise NXAPIError("Unexpected value encountered processing response.")
            new_response.append(response_dict)

        return new_response

    def _error_check(self, command_response):
        error = command_response.get("error")
        if error:
            command = command_response.get("command")
            if "data" in error:
                raise NXAPICommandError(command, error["data"]["msg"])
            else:
                raise NXAPICommandError(command, "Invalid command.")


class XMLClient(RPCBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.headers = {"content-type": "application/xml"}
        self.api = "xml"
        self.cmd_method = "cli_show"
        self.cmd_method_conf = "cli_conf"
        self.cmd_method_raw = "cli_show_ascii"

    def _nxapi_command(self, commands, method=None):
        """Send a command down the NX-API channel."""
        if method is None:
            method = self.cmd_method
        if isinstance(commands, string_types):
            commands = [commands]

        response = self._send_request(commands, method=method)
        api_response = self._process_api_response(response, commands)

        for command_response in api_response:
            self._error_check(command_response)
        return api_response

    def _nxapi_command_conf(self, commands, method=None):
        if method is None:
            method = self.cmd_method_conf
        return self._nxapi_command(commands=commands, method=method)

    def _build_payload(self, commands, method, xml_version="1.0", version="1.0"):
        xml_commands = ""
        for command in commands:
            if not xml_commands:
                # initial command is just the command itself
                xml_commands += command
            else:
                # subsequent commands are separate by semi-colon
                xml_commands += " ;{}".format(command)

        payload = """<?xml version="{xml_version}"?>
            <ins_api>
                <version>{version}</version>
                <type>{method}</type>
                <chunk>0</chunk>
                <sid>sid</sid>
                <input>{command}</input>
                <output_format>xml</output_format>
            </ins_api>""".format(
            xml_version=xml_version,
            version=version,
            method=method,
            command=xml_commands,
        )
        return payload

    def _process_api_response(self, response, commands, raw_text=False):
        xml_root = etree.fromstring(response)
        response_list = xml_root.xpath("outputs/output")
        if len(commands) != len(response_list):
            raise NXAPIXMLError(
                "XML response doesn't match expected number of commands."
            )

        return response_list

    def _error_check(self, command_response):
        """commmand_response will be an XML Etree object."""
        error_list = command_response.find("./clierror")
        command_obj = command_response.find("./input")
        if error_list is not None:
            command = command_obj.text if command_obj is not None else "Unknown command"
            msg = etree.tostring(error_list).decode()
            raise NXAPICommandError(command, msg)



# The contents of this file are licensed under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
import inspect
from netmiko import BaseConnection


def netmiko_args(optional_args):
    """Check for Netmiko arguments that were passed in as NAPALM optional arguments.

    Return a dictionary of these optional args  that will be passed into the Netmiko
    ConnectHandler call.
    """
    fields = inspect.getfullargspec(BaseConnection.__init__)
    args = fields[0]
    defaults = fields[3]

    check_self = args.pop(0)
    if check_self != "self":
        raise ValueError("Error processing Netmiko arguments")

    netmiko_argument_map = dict(zip(args, defaults))

    # Netmiko arguments that are integrated into NAPALM already
    netmiko_filter = ["ip", "host", "username", "password", "device_type", "timeout"]

    # Filter out all of the arguments that are integrated into NAPALM
    for k in netmiko_filter:
        netmiko_argument_map.pop(k)

    # Check if any of these arguments were passed in as NAPALM optional_args
    netmiko_optional_args = {}
    for k, v in netmiko_argument_map.items():
        try:
            netmiko_optional_args[k] = optional_args[k]
        except KeyError:
            pass

    # Return these arguments for use with establishing Netmiko SSH connection
    return netmiko_optional_args



# -*- coding: utf-8 -*-
"""
NAPALM CLI Tools: helpers
=========================

Defines helpers for the CLI tools.
"""
# stdlib
import ast
import sys
import logging
import getpass
import argparse
import warnings


def warning():
    warnings.simplefilter("always", DeprecationWarning)
    warnings.warn(
        "This tool has been deprecated, please use `napalm` instead\n",
        DeprecationWarning,
    )


def build_help(connect_test=False, validate=False, configure=False, napalm_cli=False):
    parser = argparse.ArgumentParser(
        description="Command line tool to handle configuration on devices using NAPALM."
        "The script will print the diff on the screen",
        epilog="Automate all the things!!!",
    )
    parser.add_argument(
        dest="hostname",
        action="store",
        help="Host where you want to deploy the configuration.",
    )
    parser.add_argument(
        "--user",
        "-u",
        dest="user",
        action="store",
        default=getpass.getuser(),
        help="User for authenticating to the host. Default: user running the script.",
    )
    parser.add_argument(
        "--password",
        "-p",
        dest="password",
        action="store",
        help="Password for authenticating to the host."
        "If you do not provide a password in the CLI you will be prompted.",
    )
    parser.add_argument(
        "--vendor",
        "-v",
        dest="vendor",
        action="store",
        required=True,
        help="Host Operating System.",
    )
    parser.add_argument(
        "--optional_args",
        "-o",
        dest="optional_args",
        action="store",
        help="String with comma separated key=value pairs passed via optional_args to the driver.",
    )
    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        help="Enables debug mode; more verbosity.",
    )

    if configure:
        parser.add_argument(
            "--strategy",
            "-s",
            dest="strategy",
            action="store",
            choices=["replace", "merge"],
            default="replace",
            help="Strategy to use to deploy configuration. Default: replace.",
        )
        parser.add_argument(
            dest="config_file",
            action="store",
            help="File containing the configuration you want to deploy.",
        )
        parser.add_argument(
            "--dry-run",
            "-d",
            dest="dry_run",
            action="store_true",
            default=None,
            help="Only returns diff, it does not deploy the configuration.",
        )
    elif validate:
        parser.add_argument(
            "--validation_file",
            "-f",
            dest="validation_file",
            action="store",
            help="Validation file containing resources derised states",
        )
    args = parser.parse_args()

    if args.password is None:
        password = getpass.getpass("Enter password: ")
        setattr(args, "password", password)

    return args


def configure_logging(logger, debug):
    if debug:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    ch = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    return logger


def parse_optional_args(optional_args):
    if optional_args is not None:
        return {
            x.split("=")[0]: ast.literal_eval(x.split("=")[1])
            for x in optional_args.split(",")
        }
    return {}
